#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

from lib_not_dr.logger import LogLevel, COLOR_SUPPORT
from lib_not_dr.logger.formatter import BaseFormatter
from lib_not_dr.logger.structure import FormattingMessage

__all__ = [
    'BaseColorFormatter',

    'LevelColorFormatter',
    'LoggerColorFormatter',
    'TimeColorFormatter',
    'TraceColorFormatter',
    'MessageColorFormatter',

    'RESET_COLOR'
]

RESET_COLOR = '\033[0m'


class BaseColorFormatter(BaseFormatter):
    name = 'BaseColorFormatter'
    # TODO 迁移老 logger 颜色
    color = {
        # Notset: just black
        LogLevel.notset: '',
        # Trace: blue
        LogLevel.trace:  '\033[38;2;138;173;244m',
        # Fine: green
        LogLevel.fine:   '\033[0;32m',
        # Debug: cyan
        LogLevel.debug:  '\033[0;36m',
        # Info: white
        LogLevel.info:   '\033[0;37m',
        # Warn: yellow
        LogLevel.warn:   '\033[0;33m',
        # Error: red
        LogLevel.error:  '\033[0;31m',
        # Fatal: red background
        LogLevel.fatal:  '\033[0;41m'
    }

    def get_color(self, message: FormattingMessage) -> str:
        for level in self.color:
            if message[0].level <= level:
                break
        else:
            level = 90
        return self.color[level]


class LevelColorFormatter(BaseColorFormatter):
    name = 'LevelColorFormatter'
    # TODO 迁移老 logger 颜色
    color = {
        # Notset: just black
        LogLevel.notset: '',
        # Trace: blue
        LogLevel.trace:  '\033[38;2;138;173;244m',
        # Fine: green
        LogLevel.fine:   '\033[35;48;2;44;44;54m',
        # Debug: cyan
        LogLevel.debug:  '\033[38;2;133;138;149m',
        # Info: white
        LogLevel.info:   '\033[0;37m',
        # Warn: yellow
        LogLevel.warn:   '\033[0;33m',
        # Error: red
        LogLevel.error:  '\033[0;31m',
        # Fatal: red background
        LogLevel.fatal:  '\033[0;41m'
    }

    @classmethod
    def _info(cls) -> str:
        return cls.add_info('colored level', 'level', 'A colored level')

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        if isinstance(message[1].get('level'), int) or not COLOR_SUPPORT:
            return message
        # 获取颜色
        color = self.get_color(message)
        # 添加颜色
        if color == '' or color == RESET_COLOR:
            return message
        message[1]['level'] = f'{color}{message[1]["level"]}{RESET_COLOR}'
        return message


class LoggerColorFormatter(BaseColorFormatter):
    name = 'LoggerColorFormatter'
    # TODO 迁移老 logger 颜色
    color = {
        # Notset: just black
        LogLevel.notset: '',
        # Trace: blue
        LogLevel.trace:  '\033[38;2;138;173;244m',
        # Fine: green
        LogLevel.fine:   '\033[0;32m',
        # Debug: cyan
        LogLevel.debug:  '\033[0;36m',
        # Info: white
        LogLevel.info:   '\033[0;37m',
        # Warn: yellow
        LogLevel.warn:   '\033[0;33m',
        # Error: red
        LogLevel.error:  '\033[0;31m',
        # Fatal: red background
        LogLevel.fatal:  '\033[38;2;245;189;230m',
    }

    @classmethod
    def _info(cls) -> str:
        return cls.add_info('colored logger name', 'logger name', 'A colored logger name')

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        if message[1].get('logger_name') is None or not COLOR_SUPPORT:
            return message
        # 获取颜色
        color = self.get_color(message)
        # 添加颜色
        if color == '' or color == RESET_COLOR:
            return message
        message[1]['logger_name'] = f'{color}{message[1]["logger_name"]}{RESET_COLOR}'
        if message[1].get('logger_tag') is not None and message[1].get('logger_tag') != '   ':
            message[1]['logger_tag'] = f'{color}{message[1]["logger_tag"]}{RESET_COLOR}'
        return message


class TimeColorFormatter(BaseColorFormatter):
    name = 'TimeColorFormatter'
    # TODO 迁移老 logger 颜色
    color = {
        # Notset: just black
        LogLevel.notset: '',
        # Trace: blue
        LogLevel.trace:  '\033[38;2;138;173;244m',
        # Fine: green
        LogLevel.fine:   '\033[0;32m',
        # Debug: cyan
        LogLevel.debug:  '\033[0;36m',
        # Info: white
        LogLevel.info:   '\033[0;37m',
        # Warn: yellow
        LogLevel.warn:   '\033[0;33m',
        # Error: red
        LogLevel.error:  '\033[0;31m',
        # Fatal: red background
        LogLevel.fatal:  '\033[38;2;255;255;0;48;2;120;10;10m',
    }

    @classmethod
    def _info(cls) -> str:
        return cls.add_info('colored time', 'time', 'A colored time')

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        if message[1].get('log_time') is None or not COLOR_SUPPORT:
            return message
        # 获取颜色
        color = self.get_color(message)
        # 添加颜色
        if color == '' or color == RESET_COLOR:
            return message
        message[1]['log_time'] = f'{color}{message[1]["log_time"]}{RESET_COLOR}'
        return message


class TraceColorFormatter(BaseColorFormatter):
    name = 'TraceColorFormatter'
    # TODO 迁移老 logger 颜色
    color = {
        # Notset: just black
        LogLevel.notset: '\033[38;2;0;255;180m',
        # Trace: blue
        LogLevel.trace:  '\033[38;2;0;255;180m',
        # Fine: green
        LogLevel.fine:   '\033[38;2;0;255;180m',
        # Debug: cyan
        LogLevel.debug:  '\033[38;2;0;255;180m',
        # Info: white
        LogLevel.info:   '\033[38;2;0;255;180m',
        # Warn: yellow
        LogLevel.warn:   '\033[38;2;0;255;180m',
        # Error: red
        LogLevel.error:  '\033[38;2;0;255;180m',
        # Fatal: red background
        LogLevel.fatal:  '\033[38;2;255;255;0;48;2;120;10;10m',
    }

    @classmethod
    def _info(cls) -> str:
        info = cls.add_info('colored logging file', 'log_source', 'A colored logging file name')
        info += '\n'
        info += cls.add_info('colored logging line', 'log_line', 'A colored logging line number')
        info += '\n'
        info += cls.add_info('colored logging function', 'log_function', 'A colored logging function name')
        return info

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        if message[0].stack_trace is None or not COLOR_SUPPORT:
            return message
        # 获取颜色
        color = self.get_color(message)
        # 添加颜色
        if color == '' or color == RESET_COLOR:
            return message
        message[1]['log_source'] = f'{color}{message[1]["log_source"]}{RESET_COLOR}'
        message[1]['log_line'] = f'{color}{message[1]["log_line"]}{RESET_COLOR}'
        message[1]['log_function'] = f'{color}{message[1]["log_function"]}{RESET_COLOR}'
        return message


class MessageColorFormatter(BaseColorFormatter):
    name = 'MessageColorFormatter'

    color = {
        # Notset: just black
        LogLevel.notset: '',
        # Trace: blue
        LogLevel.trace:  '\033[38;2;138;173;244m',
        # Fine: blue
        LogLevel.fine:   '\033[38;2;138;173;244m',
        # Debug: blue
        LogLevel.debug:  '\033[38;2;138;173;244m',
        # Info: no color
        LogLevel.info:   '',
        # Warn: yellow
        LogLevel.warn:   '\033[0;33m',
        # Error: red
        LogLevel.error:  '\033[0;31m',
        # Fatal: red background
        LogLevel.fatal:  '\033[38;2;255;255;0;48;2;120;10;10m',
    }

    @classmethod
    def _info(cls) -> str:
        return cls.add_info('colored message', 'message', 'A colored message')

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        if message[1].get('messages') is None or not COLOR_SUPPORT:
            return message
        # 获取颜色
        color = self.get_color(message)
        # 添加颜色
        if color == '' or color == RESET_COLOR:
            return message
        if message[1]['messages'][-1] == '\n':
            message[1]['messages'] = f'{color}{message[1]["messages"][:-1]}{RESET_COLOR}\n'
        else:
            message[1]['messages'] = f'{color}{message[1]["messages"]}{RESET_COLOR}'
        return message



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------
import sys

from lib_not_dr.types.options import Options

COLOR_SUPPORT = True

if sys.platform == "win32":
    try:
        # https://stackoverflow.com/questions/36760127/...
        # how-to-use-the-new-support-for-ansi-escape-sequences-in-the-windows-10-console
        from ctypes import windll

        kernel32 = windll.kernel32
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)
    except OSError:  # pragma: no cover
        COLOR_SUPPORT = False


class LogLevel(Options):
    name = 'LogLevel'
    notset: int = 0
    trace: int = 5
    fine: int = 7
    debug: int = 10
    info: int = 20
    warn: int = 30
    error: int = 40
    fatal: int = 50



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

import time
import inspect
from types import FrameType
from typing import List, Optional

from lib_not_dr.logger import LogLevel
from lib_not_dr.types.options import Options
from lib_not_dr.logger.structure import LogMessage
from lib_not_dr.logger.outstream import BaseOutputStream, StdioOutputStream


class Logger(Options):
    name = 'Logger-v2'

    outputs: List[BaseOutputStream] = [StdioOutputStream()]

    log_name: str = 'root'

    enable: bool = True
    level: int = 20  # info

    def log_for(self, level: int) -> bool:
        """
        Check if logging is enabled for a specific level.

        Args:
            level (int): The logging level to check.

        Returns:
            bool: True if logging is enabled for the given level, False otherwise.
        """
        return self.enable and level >= self.level

    def add_output(self, output: BaseOutputStream) -> None:
        """
        Add an output to the list of outputs.

        Args:
            output (BaseOutputStream): The output to be added.

        Returns:
            None
        """
        self.outputs.append(output)
        self.level = min(self.level, output.level)

    def remove_output(self, output: BaseOutputStream) -> None:
        """
        Removes the specified output from the list of outputs.

        Args:
            output (BaseOutputStream): The output to be removed.

        Returns:
            None
        """
        self.outputs.remove(output)
        self.level = max(self.level, *[output.level for output in self.outputs])

    @property
    def global_level(self) -> int:
        """
        Get the global logging level.

        Returns:
            int: The global logging level.
        """
        return self.level

    @global_level.setter
    def global_level(self, level: int) -> None:
        """
        Set the global logging level.

        Args:
            level (int): The global logging level.

        Returns:
            None
        """
        self.level = level
        for output in self.outputs:
            output.level = level

    def make_log(self,
                 messages: List[str],
                 tag: Optional[str] = None,
                 end: str = '\n',
                 split: str = ' ',
                 flush: bool = True,
                 level: int = 20,  # info
                 # log_time: Optional[float] = None,
                 # logger_name: str = 'root',
                 # logger_tag: Optional[str] = None,
                 stack_trace: Optional[FrameType] = None) -> None:
        # 检查是否需要记录
        if not self.log_for(level):
            return
        log_time = time.time_ns()
        # 处理堆栈信息
        if stack_trace is None:
            # 尝试获取堆栈信息
            if (stack := inspect.currentframe()) is not None:
                # 如果可能 尝试获取上两层的堆栈信息
                if (up_stack := stack.f_back) is not None:
                    if (upper_stack := up_stack.f_back) is not None:
                        stack_trace = upper_stack
                    else:
                        stack_trace = up_stack
                else:
                    stack_trace = stack

        message = LogMessage(messages=messages,
                             end=end,
                             split=split,
                             flush=flush,
                             level=level,
                             log_time=log_time,
                             logger_name=self.log_name,
                             logger_tag=tag,
                             stack_trace=stack_trace)
        if level >= 30:  # WARN
            for output in self.outputs:
                output.write_stderr(message)
        else:
            for output in self.outputs:
                output.write_stdout(message)
        # done?
        # 20231106 00:06

    @staticmethod
    def get_logger_by_name(name: str) -> 'Logger':
        """
        Get a logger by name.

        Args:
            name (str): The name of the logger.

        Returns:
            Logger: The logger with the specified name.
        """
        return Logger(log_name=name)

    def info(self,
             *message,
             tag: Optional[str] = None,
             end: str = '\n',
             split: str = ' ',
             flush: bool = True,
             stack_trace: Optional[FrameType] = None) -> None:
        if not self.log_for(LogLevel.info):
            return
        self.make_log(messages=list(message),
                      tag=tag,
                      end=end,
                      split=split,
                      flush=flush,
                      level=LogLevel.info,
                      stack_trace=stack_trace)

    def trace(self,
              *message,
              tag: Optional[str] = None,
              end: str = '\n',
              split: str = ' ',
              flush: bool = True,
              stack_trace: Optional[FrameType] = None) -> None:
        if not self.log_for(LogLevel.trace):
            return
        self.make_log(messages=list(message),
                      tag=tag,
                      end=end,
                      split=split,
                      flush=flush,
                      level=LogLevel.trace,
                      stack_trace=stack_trace)

    def fine(self,
             *message,
             tag: Optional[str] = None,
             end: str = '\n',
             split: str = ' ',
             flush: bool = True,
             stack_trace: Optional[FrameType] = None) -> None:
        if not self.log_for(LogLevel.fine):
            return
        self.make_log(messages=list(message),
                      tag=tag,
                      end=end,
                      split=split,
                      flush=flush,
                      level=LogLevel.fine,
                      stack_trace=stack_trace)

    def debug(self,
              *message,
              tag: Optional[str] = None,
              end: str = '\n',
              split: str = ' ',
              flush: bool = True,
              stack_trace: Optional[FrameType] = None) -> None:
        if not self.log_for(LogLevel.debug):
            return
        self.make_log(messages=list(message),
                      tag=tag,
                      end=end,
                      split=split,
                      flush=flush,
                      level=LogLevel.debug,
                      stack_trace=stack_trace)

    def warn(self,
             *message,
             tag: Optional[str] = None,
             end: str = '\n',
             split: str = ' ',
             flush: bool = True,
             stack_trace: Optional[FrameType] = None) -> None:
        if not self.log_for(LogLevel.warn):
            return
        self.make_log(messages=list(message),
                      tag=tag,
                      end=end,
                      split=split,
                      flush=flush,
                      level=LogLevel.warn,
                      stack_trace=stack_trace)

    def error(self,
              *message,
              tag: Optional[str] = None,
              end: str = '\n',
              split: str = ' ',
              flush: bool = True,
              stack_trace: Optional[FrameType] = None) -> None:
        if not self.log_for(LogLevel.error):
            return
        self.make_log(messages=list(message),
                      tag=tag,
                      end=end,
                      split=split,
                      flush=flush,
                      level=LogLevel.error,
                      stack_trace=stack_trace)

    def fatal(self,
              *message,
              tag: Optional[str] = None,
              end: str = '\n',
              split: str = ' ',
              flush: bool = True,
              stack_trace: Optional[FrameType] = None) -> None:
        if not self.log_for(LogLevel.fatal):
            return
        self.make_log(messages=list(message),
                      tag=tag,
                      end=end,
                      split=split,
                      flush=flush,
                      level=LogLevel.fatal,
                      stack_trace=stack_trace)



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from lib_not_dr.nuitka import (
        reader,
        compile
    )

__all__ = [
    "reader",
    "compile"
]



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

import time

from pathlib import Path
from types import FrameType
from typing import List, Optional, Tuple, Dict, Union

from lib_not_dr.types.options import Options

__all__ = ['LogMessage',
           'FormattingMessage']


class LogMessage(Options):
    name = 'LogMessage'

    # 消息内容本身的属性
    messages: List[str] = []
    end: str = '\n'
    split: str = ' '

    # 消息的属性
    flush: bool = None
    level: int = 20
    log_time: float = None  # time.time_ns()
    logger_name: str = 'root'
    logger_tag: Optional[str] = None
    stack_trace: Optional[FrameType] = None

    def __init__(self,
                 messages: Optional[List[str]] = None,
                 end: Optional[str] = '\n',
                 split: Optional[str] = ' ',
                 flush: Optional[bool] = None,
                 level: Optional[int] = 20,
                 log_time: Optional[float] = None,
                 logger_name: Optional[str] = 'root',
                 logger_tag: Optional[str] = None,
                 stack_trace: Optional[FrameType] = None,
                 **kwargs) -> None:
        """
        Init for LogMessage
        :param messages: message list for log
        :param end: end of message
        :param split: split for messages
        :param flush: do flush or not
        :param level: level of message
        :param log_time: time of message (default: time.time_ns())
        :param logger_name: name of logger
        :param logger_tag: tag of logger
        :param stack_trace: stack trace of logger
        :param kwargs: other options
        """
        # 为了方便使用 单独覆盖了 __init__ 方法来提供代码补全的选项
        if messages is None:
            messages = []
        super().__init__(messages=messages,
                         end=end,
                         split=split,
                         flush=flush,
                         level=level,
                         log_time=log_time,
                         logger_name=logger_name,
                         logger_tag=logger_tag,
                         stack_trace=stack_trace,
                         **kwargs)

    def init(self, **kwargs) -> bool:
        if self.log_time is None:
            self.log_time = time.time_ns()
        if not isinstance(self.flush, bool) and self.flush is not None:
            self.flush = True if self.flush else False
        return False

    def format_message(self) -> str:
        if self.split is None:
            self.split = ' '
        return self.split.join(self.messages) + self.end

    def format_for_message(self) -> Dict[str, str]:
        basic_info = self.option()

        if self.logger_tag is None:
            basic_info['logger_tag'] = '   '

        basic_info['messages'] = self.format_message()

        return basic_info

    @property
    def create_msec_3(self) -> int:
        return int(self.log_time / 1000000) % 1000


FormattingMessage = Tuple[LogMessage, Dict[str, Union[str, Path]]]

if __name__ == '__main__':
    print(LogMessage().as_markdown())



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------

__version__ = '0.2.0-rc.10'

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from lib_not_dr import (
        logger,
        nuitka,
        types,
        command
    )

__all__ = [
    '__version__',
    'logger',
    'nuitka',
    'types',
    'command'
]



#  -------------------------------
#  Difficult Rocket
#  Copyright © 2020-2023 by shenjackyuanjie 3695888@qq.com
#  All rights reserved
#  -------------------------------
import time

from pathlib import Path
from string import Template
from typing import List, Union, Optional, TYPE_CHECKING

from lib_not_dr.logger import LogLevel
from lib_not_dr.types.options import Options
from lib_not_dr.logger.structure import LogMessage, FormattingMessage

if TYPE_CHECKING:
    from lib_not_dr.logger.formatter.colors import BaseColorFormatter

__all__ = [
    'BaseFormatter',
    'MainFormatter',
    'StdFormatter'
]


class BaseFormatter(Options):
    name = 'BaseFormatter'

    sub_formatter: List['BaseFormatter'] = []
    color_formatters: List['BaseColorFormatter'] = []
    default_template: str = '[${log_time}][${level}]|${logger_name}:${logger_tag}|${messages}'

    @classmethod
    def add_info(cls, match: str, to: str, description: str) -> str:
        return f'- {to} -> ${{{match}}} : {description}'

    @classmethod
    def info(cls) -> str:
        infos = {BaseFormatter.name: BaseFormatter._info()}
        cache = ''
        for formatter in cls.sub_formatter:
            infos[formatter.name] = formatter._info()
        infos[cls.name] = cls._info()
        for name, info in infos.items():
            cache += f"## {name}\n"
            cache += info
            cache += '\n'
        return cache

    @classmethod
    def _info(cls) -> str:
        info = cls.add_info('logger_name', 'logger name', 'The name of the logger')
        info += '\n'
        info += cls.add_info('logger_tag', 'logger tag', 'The tag of the logger')
        return info

    def format_message(self,
                       message: LogMessage,
                       template: Optional[Union[Template, str]] = None) -> str:
        """
        Format message
        :param message: 输入的消息
        :param template: 日志输出模板
        :return:
        """
        basic_info = message.format_for_message()
        message, info = self._format((message, basic_info))

        if template is None:
            template = Template(self.default_template)
        elif isinstance(template, str):
            template = Template(template)

        try:
            return template.substitute(**info)
        except (KeyError, ValueError):
            return template.safe_substitute(**info)

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        """
        Format message
        :param message:
        :return:
        """
        for formatter in self.sub_formatter:
            message = formatter._format(message)
        return message

    @property
    def template(self) -> str:
        return self.default_template

    @template.setter
    def template(self, template: str) -> None:
        if not isinstance(template, str):
            raise TypeError(f'The template must be str, not {type(template)}')
        self.default_template = template


class MainFormatter(BaseFormatter):
    name = 'TraceFormatter'

    time_format: str = '%Y-%m-%d %H:%M:%S'
    msec_time_format: str = '{}-{:03d}'
    use_absolute_path: bool = False

    default_level: int = 20
    level_get_higher: bool = True

    level_name_map = {
        LogLevel.notset: 'NOTSET',
        LogLevel.trace:  ' TRACE',
        LogLevel.fine:   ' FINE ',
        LogLevel.debug:  ' DEBUG',
        LogLevel.info:   ' INFO ',
        LogLevel.warn:   ' WARN ',
        LogLevel.error:  'ERROR ',
        LogLevel.fatal:  'FATAL ',
    }
    name_level_map = {
        'NOTSET': LogLevel.notset,
        ' TRACE': LogLevel.trace,
        ' FINE ': LogLevel.fine,
        ' DEBUG': LogLevel.debug,
        ' INFO ': LogLevel.info,
        ' WARN ': LogLevel.warn,
        'ERROR ': LogLevel.error,
        'FATAL ': LogLevel.fatal,
    }

    @classmethod
    def _info(cls) -> str:
        info = cls.add_info('log_time', 'formatted time when logging', 'The time format string'
                                                                       '. See https://docs.python.org/3/library/time'
                                                                       '.html#time.strftime for more information.')
        info += '\n'
        info += cls.add_info('level', 'log level', 'The log level')
        info += '\n'
        info += cls.add_info('log_source', 'logging file', 'the logging file name')
        info += '\n'
        info += cls.add_info('log_line', 'logging line', 'the logging line number')
        info += '\n'
        info += cls.add_info('log_function', 'logging function', 'the logging function name')
        return info

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        message = self._time_format(message)
        message = self._trace_format(message)
        message = self._level_format(message)
        return message

    def _level_format(self, message: FormattingMessage) -> FormattingMessage:
        if message[0].level in self.name_level_map:
            level_tag = self.level_name_map[message[0].level]
        else:
            if self.level_get_higher:
                for level in self.name_level_map:
                    if message[0].level <= self.name_level_map[level]:
                        level_tag = level
                        break
                else:
                    level_tag = 'FATAL'
            else:
                for level in self.name_level_map:
                    if message[0].level >= self.name_level_map[level]:
                        level_tag = level
                        break
                else:
                    level_tag = 'NOTSET'
        message[1]['level'] = level_tag
        return message

    def _time_format(self, message: FormattingMessage) -> FormattingMessage:
        time_mark = time.localtime(message[0].log_time / 1000000000)
        if self.msec_time_format:
            time_mark = self.msec_time_format.format(time.strftime(self.time_format, time_mark),
                                                     message[0].create_msec_3)
        message[1]['log_time'] = time_mark
        return message

    def _trace_format(self, message: FormattingMessage) -> FormattingMessage:
        if message[0].stack_trace is None:
            return message
        path = Path(message[0].stack_trace.f_code.co_filename)
        if self.use_absolute_path:
            message[1]['log_source'] = path.absolute()
        message[1]['log_source'] = path
        message[1]['log_line'] = message[0].stack_trace.f_lineno
        message[1]['log_function'] = message[0].stack_trace.f_code.co_name
        return message


class StdFormatter(BaseFormatter):
    name = 'StdFormatter'

    enable_color: bool = True

    sub_formatter: List[BaseFormatter] = []
    from lib_not_dr.logger.formatter.colors import (LevelColorFormatter,
                                                    LoggerColorFormatter,
                                                    TimeColorFormatter,
                                                    TraceColorFormatter,
                                                    MessageColorFormatter)
    color_formatters: List[BaseFormatter] = [LevelColorFormatter(),
                                             LoggerColorFormatter(),
                                             TimeColorFormatter(),
                                             TraceColorFormatter(),
                                             MessageColorFormatter()]

    def __init__(self,
                 enable_color: bool = True,
                 sub_formatter: Optional[List[BaseFormatter]] = None,
                 main_formatter: Optional[MainFormatter] = None,
                 color_formatters: Optional[List[BaseFormatter]] = None,
                 **kwargs) -> None:
        """
        Initialize the StdFormatter
        :param enable_color: enable color
        :param sub_formatter: list of sub formatter
        :param color_formatters: list of color formatter
        :param kwargs: other options
        """
        # 同 structures.LogMessage.__init__ 的注释 (逃)
        self.enable_color = enable_color
        if sub_formatter is not None:
            self.sub_formatter = sub_formatter
        if color_formatters is not None:
            self.color_formatters = color_formatters
        if main_formatter is not None and isinstance(main_formatter, MainFormatter):
            self.main_formatter = main_formatter
        else:
            self.main_formatter = MainFormatter()
        super().__init__(**kwargs)

    def _format(self, message: FormattingMessage) -> FormattingMessage:
        super()._format(message)

        if not self.enable_color:
            return message

        for formatter in self.color_formatters:
            message = formatter._format(message)

        return message

    @classmethod
    def _info(cls) -> str:
        return 'None'


if __name__ == '__main__':
    import inspect

    log_message = LogMessage(messages=['Hello World!'],
                             level=7,
                             stack_trace=inspect.currentframe(),
                             logger_tag='tester',
                             logger_name='test')

    print(StdFormatter.info())
    print(StdFormatter().format_message(log_message))

    std_format = StdFormatter()
    std_format.default_template = "${log_time}|${logger_name}|${logger_tag}|${log_source}:${log_line}|${log_function}|${level}|${messages}"

    test_levels = (0, 5, 7, 10, 20, 30, 40, 50)

    print("with color")

    for test_level in test_levels:
        log_message.level = test_level
        print(std_format.format_message(log_message), end='')

    print("without color")

    std_format.enable_color = False

    for test_level in test_levels:
        log_message.level = test_level
        print(std_format.format_message(log_message), end='')

    print(std_format.as_markdown())
